"""Command-line interface for bmi-topography """
import click

from .topography import Topography


@click.command()
@click.version_option()
@click.option("-q", "--quiet", is_flag=True, help="Enables quiet mode.")
@click.option(
    "--dem-type",
    type=click.Choice(Topography.VALID_DEM_TYPES, case_sensitive=True),
    default=Topography.DEFAULT["dem_type"],
    help="The global raster dataset.",
    show_default=True,
)
@click.option(
    "--south",
    type=click.FloatRange(-90, 90),
    default=Topography.DEFAULT["south"],
    help="WGS 84 bounding box south coordinate, in degrees, on [-90,90].",
    show_default=True,
)
@click.option(
    "--north",
    type=click.FloatRange(-90, 90),
    default=Topography.DEFAULT["north"],
    help="WGS 84 bounding box north coordinate, in degrees, on [-90,90].",
    show_default=True,
)
@click.option(
    "--west",
    type=click.FloatRange(-180, 180),
    default=Topography.DEFAULT["west"],
    help="WGS 84 bounding box west coordinate, in degrees, on [-180,180].",
    show_default=True,
)
@click.option(
    "--east",
    type=click.FloatRange(-180, 180),
    default=Topography.DEFAULT["east"],
    help="WGS 84 bounding box east coordinate, in degrees, on [-180,180].",
    show_default=True,
)
@click.option(
    "--output-format",
    type=click.Choice(Topography.VALID_OUTPUT_FORMATS.keys(), case_sensitive=True),
    default=Topography.DEFAULT["output_format"],
    help="Output file format.",
    show_default=True,
)
@click.option(
    "--api-key",
    type=str,
    help="OpenTopography API key.",
    show_default=True,
)
@click.option("--no-fetch", is_flag=True, help="Do not fetch data from server.")
def main(quiet, dem_type, south, north, west, east, output_format, api_key, no_fetch):
    """Fetch and cache NASA SRTM and JAXA ALOS land elevation data

    Some datasets require an OpenTopography API key. You can find instructions
    on how to obtain one from the OpenTopography website:

        https://opentopography.org/blog/introducing-api-keys-access-opentopography-global-datasets

    Once you have a key, you can pass it to the bmi-topography
    command in one of two ways: 1) through the environment variable
    OPENTOPOGRAPHY_API_KEY, or 2) as the contents of the file
    ".opentopography.txt" located either in your current directory or your home
    directory.
    """
    topo = Topography(dem_type, south, north, west, east, output_format)
    if not no_fetch:
        if not quiet:
            click.secho("Fetching data...", fg="yellow", err=True)
        path_to_dem = topo.fetch()
        if not quiet:
            click.secho(
                "File downloaded to {}".format(getattr(topo, "cache_dir")),
                fg="green",
                err=True,
            )
        print(path_to_dem)



from ._version import __version__
from .bbox import BoundingBox
from .bmi import BmiTopography
from .topography import Topography

__all__ = ["Topography", "BoundingBox", "BmiTopography", "__version__"]



import os
from unittest import mock

import pytest

from bmi_topography.api_key import ApiKey, _find_first_of
from bmi_topography.errors import BadApiKeySource, BadKeyError, MissingKeyError


def copy_environ(exclude=None):
    if exclude is None:
        exclude = {}
    elif isinstance(exclude, str):
        exclude = {exclude}

    return {key: value for key, value in os.environ.items() if key not in exclude}


def test_find_user_api_key_not_found():
    """The API key is not given anywhere"""
    env = copy_environ(exclude="OPENTOPOGRAPHY_API_KEY")
    with mock.patch.dict(os.environ, env, clear=True):
        with pytest.raises(MissingKeyError):
            ApiKey.from_env()


@pytest.mark.parametrize("bad_key", [None, 0, 1, ""])
def test_bad_user_key(bad_key):
    with pytest.raises(BadKeyError):
        ApiKey(bad_key)


@pytest.mark.parametrize("bad_source", [None, 0, 1, "", "not-a-source"])
def test_bad_source(bad_source):
    with pytest.raises(BadApiKeySource):
        ApiKey("foobar", source=bad_source)


@mock.patch.dict(os.environ, {"OPENTOPOGRAPHY_API_KEY": "foo"})
def test_find_user_api_key_env(tmpdir):
    """The API key is an environment variable"""
    with tmpdir.as_cwd():
        with open(".opentopography.txt", "w") as fp:
            fp.write("bar")
        key = ApiKey.from_sources()
    assert key == "foo"
    assert key.source == "env"
    assert key == ApiKey("foo", source="env")


@mock.patch.dict(os.environ, {"OPENTOPOGRAPHY_API_KEY": "foo"})
def test_find_user_api_key_from_file(tmpdir):
    """The API key is in a file"""
    env = copy_environ(exclude="OPENTOPOGRAPHY_API_KEY")
    with tmpdir.as_cwd():
        with open(".opentopography.txt", "w") as fp:
            fp.write("bar")

        with mock.patch.dict(os.environ, env, clear=True):
            key = ApiKey.from_sources()
            assert key == "bar"
            assert key.source.startswith("file:")
            assert key.source.endswith(".opentopography.txt")


@mock.patch.dict(os.environ, {"OPENTOPOGRAPHY_API_KEY": "foo"})
def test_find_user_api_key_from_user(tmpdir):
    """The API key is in a file"""
    env = copy_environ(exclude="OPENTOPOGRAPHY_API_KEY")
    with tmpdir.as_cwd():
        with mock.patch.dict(os.environ, env, clear=True):
            key = ApiKey.from_sources(api_key="foobar")
            assert key == "foobar"
            assert key.source.startswith("user")


def test_find_user_api_key_from_missing_file(tmpdir):
    """The API key is in a file"""
    env = copy_environ(exclude="OPENTOPOGRAPHY_API_KEY")
    with tmpdir.as_cwd():
        with mock.patch.dict(os.environ, env, clear=True):
            with pytest.raises(MissingKeyError):
                ApiKey.from_file()


def test_default_to_demo_key(tmpdir):
    """If a key can't be found, use the demo key"""
    env = copy_environ(exclude="OPENTOPOGRAPHY_API_KEY")
    with tmpdir.as_cwd():
        with mock.patch.dict(os.environ, env, clear=True):
            key = ApiKey.from_sources()
            assert key.is_demo_key()
            assert key.source == "demo"


def test_read_first_missing(tmpdir):
    with tmpdir.as_cwd():
        assert _find_first_of(["foo.txt"]) is None
        assert _find_first_of([]) is None


def test_read_first_file(tmpdir):
    with tmpdir.as_cwd():
        with open("foo.txt", "w") as fp:
            fp.write("foo")
        with open("bar.txt", "w") as fp:
            fp.write("bar")

        assert _find_first_of(["foo.txt", "bar.txt"]).name == "foo.txt"
        assert _find_first_of(["bar.txt", "foo.txt"]).name == "bar.txt"


def test_use_demo_key_is_a_string():
    key = ApiKey(ApiKey.DEMO_API_KEY, source="demo")
    assert len(key) > 0
    assert key.source == "demo"
    assert key.is_demo_key()


def test_use_demo_key_issues_warning():
    with pytest.warns(UserWarning):
        ApiKey(ApiKey.DEMO_API_KEY)


def test_api_key_repr():
    key = ApiKey("foobar")
    assert eval(repr(key)) == key



"""Test Topography class"""
import os

import pytest

from bmi_topography import Topography

API_URL = "https://portal.opentopography.org/API/globaldem"
CENTER_LAT = 40.0
CENTER_LON = -105.0
WIDTH = 0.01


def test_data_url():
    r = Topography.data_url()
    assert r == API_URL


def test_invalid_dem_type():
    with pytest.raises(ValueError):
        Topography(dem_type="foo", output_format=Topography.DEFAULT["output_format"])


def test_invalid_output_format():
    with pytest.raises(ValueError):
        Topography(dem_type=Topography.DEFAULT["dem_type"], output_format="foo")


def test_valid_bbox():
    topo = Topography(
        dem_type=Topography.DEFAULT["dem_type"],
        output_format=Topography.DEFAULT["output_format"],
        south=Topography.DEFAULT["south"],
        west=Topography.DEFAULT["west"],
        north=Topography.DEFAULT["north"],
        east=Topography.DEFAULT["east"],
    )
    assert topo.bbox.south == Topography.DEFAULT["south"]
    assert topo.bbox.west == Topography.DEFAULT["west"]
    assert topo.bbox.north == Topography.DEFAULT["north"]
    assert topo.bbox.east == Topography.DEFAULT["east"]
    assert topo.bbox.south < topo.bbox.north
    assert topo.bbox.west < topo.bbox.east


@pytest.mark.parametrize("cache_dir", (".", "./cache"))
def test_cache_dir(tmpdir, cache_dir):
    with tmpdir.as_cwd():
        topo = Topography(
            dem_type=Topography.DEFAULT["dem_type"],
            output_format=Topography.DEFAULT["output_format"],
            south=Topography.DEFAULT["south"],
            west=Topography.DEFAULT["west"],
            north=Topography.DEFAULT["north"],
            east=Topography.DEFAULT["east"],
            cache_dir=cache_dir,
        )
        assert topo.cache_dir.is_absolute()
        assert list(topo.cache_dir.glob("*.tif")) == []


def test_cached_data(tmpdir, shared_datadir):
    with tmpdir.as_cwd():
        Topography(
            dem_type=Topography.DEFAULT["dem_type"],
            output_format=Topography.DEFAULT["output_format"],
            south=Topography.DEFAULT["south"],
            west=Topography.DEFAULT["west"],
            north=Topography.DEFAULT["north"],
            east=Topography.DEFAULT["east"],
            cache_dir=shared_datadir,
        )
        assert len(tmpdir.listdir(fil=lambda f: f.ext == ".tif")) == 0


@pytest.mark.skipif("NO_FETCH" in os.environ, reason="NO_FETCH is set")
@pytest.mark.parametrize("dem_type", Topography.VALID_DEM_TYPES)
@pytest.mark.parametrize(
    "output_format,file_type", Topography.VALID_OUTPUT_FORMATS.items()
)
def test_fetch(tmpdir, dem_type, output_format, file_type):
    with tmpdir.as_cwd():
        topo = Topography(
            dem_type=dem_type,
            output_format=output_format,
            south=CENTER_LAT - WIDTH,
            west=CENTER_LON - WIDTH,
            north=CENTER_LAT + WIDTH,
            east=CENTER_LON + WIDTH,
            cache_dir=".",
        )
        topo.fetch()
        assert len(tmpdir.listdir(fil=lambda f: f.ext == "." + file_type)) == 1


def test_load(tmpdir, shared_datadir):
    with tmpdir.as_cwd():
        topo = Topography(
            dem_type=Topography.DEFAULT["dem_type"],
            output_format=Topography.DEFAULT["output_format"],
            south=Topography.DEFAULT["south"],
            west=Topography.DEFAULT["west"],
            north=Topography.DEFAULT["north"],
            east=Topography.DEFAULT["east"],
            cache_dir=shared_datadir,
        )
        topo.load()
        assert topo.da is not None
        assert topo.da.attrs["units"] == "degrees"



# -*- coding: utf-8 -*-
from collections import namedtuple
from typing import Tuple

import numpy
import yaml
from bmipy import Bmi

from .topography import Topography

BmiVar = namedtuple(
    "BmiVar", ["dtype", "itemsize", "nbytes", "units", "location", "grid"]
)
BmiGridUniformRectilinear = namedtuple(
    "BmiGridUniformRectilinear", ["shape", "yx_spacing", "yx_of_lower_left"]
)


class BmiTopography(Bmi):

    """BMI-mediated access to NASA SRTM land elevation data."""

    _name = "bmi-topography"
    _input_var_names = ()
    _output_var_names = ("land_surface__elevation",)

    def __init__(self) -> None:
        self._config = {}
        self._da = None
        self._grid = {}
        self._var = None

    def finalize(self) -> None:
        """Perform tear-down tasks for the model.

        Perform all tasks that take place after exiting the model's time
        loop. This typically includes deallocating memory, closing files and
        printing reports.
        """
        self._da = None

    def get_component_name(self) -> str:
        """Name of the component.

        Returns
        -------
        str
            The name of the component.
        """
        return self._name

    def get_current_time(self) -> float:
        """Current time of the model.

        Returns
        -------
        float
            The current model time.
        """
        return 0.0

    def get_end_time(self) -> float:
        """End time of the model.

        Returns
        -------
        float
            The maximum model time.
        """
        return 0.0

    def get_grid_edge_count(self, grid: int) -> int:
        """Get the number of edges in the grid.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        int
            The total number of grid edges.
        """
        raise NotImplementedError("get_grid_edge_count")

    def get_grid_edge_nodes(
        self, grid: int, edge_nodes: numpy.ndarray
    ) -> numpy.ndarray:
        """Get the edge-node connectivity.

        Parameters
        ----------
        grid : int
            A grid identifier.
        edge_nodes : ndarray of int, shape *(2 x nnodes,)*
            A numpy array to place the edge-node connectivity. For each edge,
            connectivity is given as node at edge tail, followed by node at
            edge head.

        Returns
        -------
        ndarray of int
            The input numpy array that holds the edge-node connectivity.
        """
        raise NotImplementedError("get_grid_edge_nodes")

    def get_grid_face_count(self, grid: int) -> int:
        """Get the number of faces in the grid.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        int
            The total number of grid faces.
        """
        raise NotImplementedError("get_grid_face_count")

    def get_grid_face_edges(
        self, grid: int, face_edges: numpy.ndarray
    ) -> numpy.ndarray:
        """Get the face-edge connectivity.

        Parameters
        ----------
        grid : int
            A grid identifier.
        face_edges : ndarray of int
            A numpy array to place the face-edge connectivity.

        Returns
        -------
        ndarray of int
            The input numpy array that holds the face-edge connectivity.
        """
        raise NotImplementedError("get_grid_face_edges")

    def get_grid_face_nodes(
        self, grid: int, face_nodes: numpy.ndarray
    ) -> numpy.ndarray:
        """Get the face-node connectivity.

        Parameters
        ----------
        grid : int
            A grid identifier.
        face_nodes : ndarray of int
            A numpy array to place the face-node connectivity. For each face,
            the nodes (listed in a counter-clockwise direction) that form the
            boundary of the face.

        Returns
        -------
        ndarray of int
            The input numpy array that holds the face-node connectivity.
        """
        raise NotImplementedError("get_grid_face_nodes")

    def get_grid_node_count(self, grid: int) -> int:
        """Get the number of nodes in the grid.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        int
            The total number of grid nodes.
        """
        raise NotImplementedError("get_grid_node_count")

    def get_grid_nodes_per_face(
        self, grid: int, nodes_per_face: numpy.ndarray
    ) -> numpy.ndarray:
        """Get the number of nodes for each face.

        Parameters
        ----------
        grid : int
            A grid identifier.
        nodes_per_face : ndarray of int, shape *(nfaces,)*
            A numpy array to place the number of edges per face.

        Returns
        -------
        ndarray of int
            The input numpy array that holds the number of nodes per edge.
        """
        raise NotImplementedError("get_grid_nodes_per_face")

    def get_grid_origin(self, grid: int, origin: numpy.ndarray) -> numpy.ndarray:
        """Get coordinates for the lower-left corner of the computational grid.

        Parameters
        ----------
        grid : int
            A grid identifier.
        origin : ndarray of float, shape *(ndim,)*
            A numpy array to hold the coordinates of the lower-left corner of
            the grid.

        Returns
        -------
        ndarray of float
            The input numpy array that holds the coordinates of the grid's
            lower-left corner.
        """
        origin[:] = self._grid[grid].yx_of_lower_left
        return origin

    def get_grid_rank(self, grid: int) -> int:
        """Get number of dimensions of the computational grid.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        int
            Rank of the grid.
        """
        return len(self._grid[grid].shape)

    def get_grid_shape(self, grid: int, shape: numpy.ndarray) -> numpy.ndarray:
        """Get dimensions of the computational grid.

        Parameters
        ----------
        grid : int
            A grid identifier.
        shape : ndarray of int, shape *(ndim,)*
            A numpy array into which to place the shape of the grid.

        Returns
        -------
        ndarray of int
            The input numpy array that holds the grid's shape.
        """
        shape[:] = self._grid[grid].shape
        return shape

    def get_grid_size(self, grid: int) -> int:
        """Get the total number of elements in the computational grid.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        int
            Size of the grid.
        """
        return int(numpy.prod(self._grid[grid].shape))

    def get_grid_spacing(self, grid: int, spacing: numpy.ndarray) -> numpy.ndarray:
        """Get distance between nodes of the computational grid.

        Parameters
        ----------
        grid : int
            A grid identifier.
        spacing : ndarray of float, shape *(ndim,)*
            A numpy array to hold the spacing between grid rows and columns.

        Returns
        -------
        ndarray of float
            The input numpy array that holds the grid's spacing.
        """
        spacing[:] = self._grid[grid].yx_spacing
        return spacing

    def get_grid_type(self, grid: int) -> str:
        """Get the grid type as a string.

        Parameters
        ----------
        grid : int
            A grid identifier.

        Returns
        -------
        str
            Type of grid as a string.
        """
        return "uniform_rectilinear"

    def get_grid_x(self, grid: int, x: numpy.ndarray) -> numpy.ndarray:
        """Get coordinates of grid nodes in the x direction.

        Parameters
        ----------
        grid : int
            A grid identifier.
        x : ndarray of float, shape *(nrows,)*
            A numpy array to hold the x-coordinates of the grid node columns.

        Returns
        -------
        ndarray of float
            The input numpy array that holds the grid's column x-coordinates.
        """
        x[:] = self._da.x.values
        return x

    def get_grid_y(self, grid: int, y: numpy.ndarray) -> numpy.ndarray:
        """Get coordinates of grid nodes in the y direction.

        Parameters
        ----------
        grid : int
            A grid identifier.
        y : ndarray of float, shape *(ncols,)*
            A numpy array to hold the y-coordinates of the grid node rows.

        Returns
        -------
        ndarray of float
            The input numpy array that holds the grid's row y-coordinates.
        """
        y[:] = self._da.y.values
        return y

    def get_grid_z(self, grid: int, z: numpy.ndarray) -> numpy.ndarray:
        """Get coordinates of grid nodes in the z direction.

        Parameters
        ----------
        grid : int
            A grid identifier.
        z : ndarray of float, shape *(nlayers,)*
            A numpy array to hold the z-coordinates of the grid nodes layers.

        Returns
        -------
        ndarray of float
            The input numpy array that holds the grid's layer z-coordinates.
        """
        raise NotImplementedError("get_grid_z")

    def get_input_item_count(self) -> int:
        """Count of a model's input variables.

        Returns
        -------
        int
          The number of input variables.
        """
        return len(self._input_var_names)

    def get_input_var_names(self) -> Tuple[str]:
        """List of a model's input variables.

        Input variable names must be CSDMS Standard Names, also known
        as *long variable names*.

        Returns
        -------
        list of str
            The input variables for the model.

        Notes
        -----
        Standard Names enable the CSDMS framework to determine whether
        an input variable in one model is equivalent to, or compatible
        with, an output variable in another model. This allows the
        framework to automatically connect components.

        Standard Names do not have to be used within the model.
        """
        return self._input_var_names

    def get_output_item_count(self) -> int:
        """Count of a model's output variables.

        Returns
        -------
        int
          The number of output variables.
        """
        return len(self._output_var_names)

    def get_output_var_names(self) -> Tuple[str]:
        """List of a model's output variables.

        Output variable names must be CSDMS Standard Names, also known
        as *long variable names*.

        Returns
        -------
        list of str
            The output variables for the model.
        """
        return self._output_var_names

    def get_start_time(self) -> float:
        """Start time of the model.

        Model times should be of type float.

        Returns
        -------
        float
            The model start time.
        """
        return 0.0

    def get_time_step(self) -> float:
        """Current time step of the model.

        The model time step should be of type float.

        Returns
        -------
        float
            The time step used in model.
        """
        return 0.0

    def get_time_units(self) -> str:
        """Time units of the model.

        Returns
        -------
        float
            The model time unit; e.g., `days` or `s`.

        Notes
        -----
        CSDMS uses the UDUNITS standard from Unidata.
        """
        return "1"

    def get_value(self, name: str, dest: numpy.ndarray) -> numpy.ndarray:
        """Get a copy of values of the given variable.

        This is a getter for the model, used to access the model's
        current state. It returns a *copy* of a model variable, with
        the return type, size and rank dependent on the variable.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.
        dest : ndarray
            A numpy array into which to place the values.

        Returns
        -------
        ndarray
            The same numpy array that was passed as an input buffer.
        """
        dest[:] = self.get_value_ptr(name).reshape(-1).copy()
        return dest

    def get_value_at_indices(
        self, name: str, dest: numpy.ndarray, inds: numpy.ndarray
    ) -> numpy.ndarray:
        """Get values at particular indices.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.
        dest : ndarray
            A numpy array into which to place the values.
        indices : array_like
            The indices into the variable array.

        Returns
        -------
        array_like
            Value of the model variable at the given location.
        """
        dest[:] = self.get_value_ptr(name).reshape(-1)[inds]
        return dest

    def get_value_ptr(self, name: str) -> numpy.ndarray:
        """Get a reference to values of the given variable.

        This is a getter for the model, used to access the model's
        current state. It returns a reference to a model variable,
        with the return type, size and rank dependent on the variable.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        array_like
            A reference to a model variable.
        """
        return self._da.values

    def get_var_grid(self, name: str) -> int:
        """Get grid identifier for the given variable.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        int
          The grid identifier.
        """
        return self._var.grid

    def get_var_itemsize(self, name: str) -> int:
        """Get memory use for each array element in bytes.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        int
            Item size in bytes.
        """
        return self._var.itemsize

    def get_var_location(self, name: str) -> str:
        """Get the grid element type that the a given variable is defined on.

        The grid topology can be composed of *nodes*, *edges*, and *faces*.

        *node*
            A point that has a coordinate pair or triplet: the most
            basic element of the topology.

        *edge*
            A line or curve bounded by two *nodes*.

        *face*
            A plane or surface enclosed by a set of edges. In a 2D
            horizontal application one may consider the word “polygon”,
            but in the hierarchy of elements the word “face” is most common.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        str
            The grid location on which the variable is defined. Must be one of
            `"node"`, `"edge"`, or `"face"`.

        Notes
        -----
        CSDMS uses the `ugrid conventions`_ to define unstructured grids.

        .. _ugrid conventions: http://ugrid-conventions.github.io/ugrid-conventions
        """
        return self._var.location

    def get_var_nbytes(self, name: str) -> int:
        """Get size, in bytes, of the given variable.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        int
            The size of the variable, counted in bytes.
        """
        return self._var.nbytes

    def get_var_type(self, name: str) -> str:
        """Get data type of the given variable.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        str
            The Python variable type; e.g., ``str``, ``int``, ``float``.
        """
        return self._var.dtype

    def get_var_units(self, name: str) -> str:
        """Get units of the given variable.

        Standard unit names, in lower case, should be used, such as
        ``meters`` or ``seconds``. Standard abbreviations, like ``m`` for
        meters, are also supported. For variables with compound units,
        each unit name is separated by a single space, with exponents
        other than 1 placed immediately after the name, as in ``m s-1``
        for velocity, ``W m-2`` for an energy flux, or ``km2`` for an
        area.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.

        Returns
        -------
        str
            The variable units.

        Notes
        -----
        CSDMS uses the `UDUNITS`_ standard from Unidata.

        .. _UDUNITS: http://www.unidata.ucar.edu/software/udunits
        """
        return self._var.units

    def initialize(self, config_file: str) -> None:
        """Perform startup tasks for the model.

        Perform all tasks that take place before entering the model's time
        loop, including opening files and initializing the model state. Model
        inputs are read from a text-based configuration file, specified by
        `filename`.

        Parameters
        ----------
        config_file : str, optional
            The path to the model configuration file.

        Notes
        -----
        Models should be refactored, if necessary, to use a
        configuration file. CSDMS does not impose any constraint on
        how configuration files are formatted, although YAML is
        recommended. A template of a model's configuration file
        with placeholder values is used by the BMI.
        """
        if config_file:
            with open(config_file, "r") as fp:
                self._config = yaml.safe_load(fp).get("bmi-topography", {})
        else:
            self._config = Topography.DEFAULT.copy()
        self._da = Topography(**self._config).load()

        self._grid = {
            0: BmiGridUniformRectilinear(
                shape=self._da.rio.shape,
                yx_spacing=(
                    abs(self._da.rio.transform().e),
                    abs(self._da.rio.transform().a),
                ),
                yx_of_lower_left=(
                    float(self._da.y.min().data),
                    float(self._da.x.min().data),
                ),
            )
        }

        self._var = BmiVar(
            dtype=str(self._da.values.dtype),
            itemsize=self._da.values.itemsize,
            nbytes=self._da.values.nbytes,
            location="face",
            units=self._da.attrs["units"],
            grid=0,
        )

    def set_value(self, name: str, values: numpy.ndarray) -> None:
        """Specify a new value for a model variable.

        This is the setter for the model, used to change the model's
        current state. It accepts, through *src*, a new value for a
        model variable, with the type, size and rank of *src*
        dependent on the variable.

        Parameters
        ----------
        var_name : str
            An input or output variable name, a CSDMS Standard Name.
        src : array_like
            The new value for the specified variable.
        """
        raise NotImplementedError("set_value")

    def set_value_at_indices(
        self, name: str, inds: numpy.ndarray, src: numpy.ndarray
    ) -> None:
        """Specify a new value for a model variable at particular indices.

        Parameters
        ----------
        var_name : str
            An input or output variable name, a CSDMS Standard Name.
        indices : array_like
            The indices into the variable array.
        src : array_like
            The new value for the specified variable.
        """
        raise NotImplementedError("set_value_at_indices")

    def update(self) -> None:
        """Advance model state by one time step.

        Perform all tasks that take place within one pass through the model's
        time loop. This typically includes incrementing all of the model's
        state variables. If the model's state variables don't change in time,
        then they can be computed by the :func:`initialize` method and this
        method can return with no action.
        """
        raise NotImplementedError("update")

    def update_until(self, time: float) -> None:
        """Advance model state until the given time.

        Parameters
        ----------
        time : float
            A model time later than the current model time.
        """
        raise NotImplementedError("update_until")



"""Test bmi-topography command-line interface"""
import pathlib

from click.testing import CliRunner

from bmi_topography.cli import main


def test_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output


def test_version():
    runner = CliRunner()
    result = runner.invoke(main, ["--version"])
    assert result.exit_code == 0
    assert "version" in result.output


def test_defaults():
    runner = CliRunner()
    result = runner.invoke(main, ["--quiet"])
    assert pathlib.Path(result.output.strip()).is_file()
    assert result.exit_code == 0


def test_quiet():
    runner = CliRunner()
    quiet_lines = runner.invoke(main, ["--quiet"]).output.splitlines()
    verbose_lines = runner.invoke(main).output.splitlines()

    assert len(verbose_lines) > len(quiet_lines)


def test_demtype_valid():
    runner = CliRunner()
    result = runner.invoke(main, ["--dem-type=SRTMGL1", "--no-fetch"])
    assert result.exit_code == 0


def test_demtype_invalid():
    runner = CliRunner()
    result = runner.invoke(main, ["--dem-type=foobar"])
    assert result.exit_code != 0


def test_demtype_is_case_sensitive():
    runner = CliRunner()
    result = runner.invoke(main, ["--dem-type=srtmgl1"])
    assert result.exit_code != 0


def test_south_inrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--south=-90.0", "--no-fetch"])
    assert result.exit_code == 0


def test_south_outrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--south=200.0"])
    assert result.exit_code != 0


def test_north_inrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--north=90.0", "--no-fetch"])
    assert result.exit_code == 0


def test_north_outrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--north=200.0"])
    assert result.exit_code != 0


def test_west_inrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--west=-180.0", "--no-fetch"])
    assert result.exit_code == 0


def test_west_outrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--west=400.0"])
    assert result.exit_code != 0


def test_east_inrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--east=180.0", "--no-fetch"])
    assert result.exit_code == 0


def test_east_outrange():
    runner = CliRunner()
    result = runner.invoke(main, ["--east=400.0"])
    assert result.exit_code != 0


def test_output_format_valid():
    runner = CliRunner()
    result = runner.invoke(main, ["--output-format=GTiff", "--no-fetch"])
    assert result.exit_code == 0


def test_output_format_invalid():
    runner = CliRunner()
    result = runner.invoke(main, ["--output-format=foobar"])
    assert result.exit_code != 0


def test_output_format_is_case_sensitive():
    runner = CliRunner()
    result = runner.invoke(main, ["--output-format=gtiff"])
    assert result.exit_code != 0



"""An example used in the README and documentation."""
import matplotlib.pyplot as plt

from bmi_topography import Topography

params = Topography.DEFAULT.copy()
params["south"] = 39.93
params["north"] = 40.00
params["west"] = -105.33
params["east"] = -105.26

boulder = Topography(**params)

boulder.fetch()
boulder.load()

boulder.da.spatial_ref

boulder.da.plot()
plt.show()



class BmiTopographyError(Exception):

    """Base exception for all bmi_topography errors."""

    pass


class MissingKeyError(BmiTopographyError):

    """Raise if an API key cannot be found."""

    pass


class BadKeyError(BmiTopographyError):

    """Raise for an invalid key."""

    pass


class BadApiKeySource(BmiTopographyError):

    """Raise for an invalid API key source."""

    pass
